# This file is part of the "FDN tools" project.
# You can use and redistribute it under the terms of the
# GNU General Public License.
# (c) French Data Network 1997-2005
# Authors and contributors:
# - Benjamin Bayart
# - Thomas Parmelan
# - Stephane Jourdois
package FDN::Common::Config;

use strict;
use Exporter;

our @ISA = qw ( Exporter );

our @EXPORT = qw (
	Config_Files_List
	Load_Config
	Unload_Config
	Config_Sections_List
	Config_Section
	Config_Section_Keys
	Config_Key

	Rewrite_Table
	Rewrite_Props
	
	Create_Config
	Write_Config 

	Forget_Config

	Config_Set_Var
);

use warnings;

use Storable;
use Encode;
use FDN::Common::Logger;
use FDN::Common::Modules;
use FDN::Common::Utils qw/slurp_file/;



##############################################################################
#
#   Variables globales
#

our $CONFIG_ROOT = '/etc/fdn/';	# Racine de l'arborescence de configuration
our $debug = 0;	# Par dfaut pas de debug : beaucoup trop bavard

our %files = ();

##############################################################################
#
#   Fonctions internes
#

sub __debug($) {
    my ($msg) = @_;

    debug $msg
	if $debug and defined $msg;
}


sub __effective_filename($) {
    my ($filename) = @_;
    
    if ($filename =~ m{^/}) {
	# Chemin absolu: on l'utilise tel quel.
	__debug '__effective_filename: absolute pathname used';
    } elsif ($filename =~ m{^\./}) {
	# Path relatif au repertoire courant: warnning et utilise tel quel.
	warning "__effective_filename: relative pathname used '".$filename."'";
    } else {
	__debug '__effective_filename: CONFIG_ROOT used';
	$filename = $CONFIG_ROOT . $filename;
    }

    return $filename;
}

# Avec le Perl de Woody, File::stat fait un warning lors d'un stat sur un
# fichier inexistant. Ce warning n'est pas dbrayable mme avec un eval
# contenant un no warnings 'all' si on met un -w en ligne de commande ou sur la
# ligne shebang. Avec le Perl de Sid, il n'y a pas de warning et c'est trs
# bien comme a !
sub __stat ($) {
    my ($filename) = @_;

    if ($^V and $^V lt v5.8.0) {
	# __debug "__stat: Perl obsolte, on fait un workaround pour File::stat";
	if (-f $filename) {
	    use File::stat;
	    return stat $filename;
	}
	else {
	    return undef;
	}
    }
    else {
	# Perl >= 5.8
	# __debug "Perl rcent";
	use File::stat;
	return stat $filename;
    }
}


sub __parse_file ($$$) {
    my ($filename, $binfile, $parse_func) = @_;
    my $ret;

    return undef
	unless defined $filename  and defined $parse_func;

   unless ($files{$filename}{condition_variable}) {
	return undef unless defined $binfile;
	__debug '__parse_file ' . $filename . ' -> ' . $binfile;

	$ret = &{$parse_func}($filename);

	# le fichier .bin doit tre en rw pour le groupe sitacfg
	my $old_umask = umask 0007;
	__debug '  store ' . $binfile;
	eval { store $ret, $binfile; };

	if ($@) {
	    chomp $@;
	    warning "  store: $@"; # ce n'est pas une erreur bloquante
	}
	umask $old_umask;

    }
    else  {
	__debug '__parse_file (conditional parse)' . $filename . ' -> no binfile';
	require File::Temp;
  
	my $cfgfiletmp = File::Temp::tempnam("/tmp","");
	__condition_parse ($filename, $cfgfiletmp, $files{$filename}{condition_variable});

	$ret = &{$parse_func}($cfgfiletmp);
	unlink $cfgfiletmp;
    }

    return $ret;
}


sub __get_file ($$) {
    my ($filename, $parse_func) = @_;

    return undef
	unless defined $filename and defined $parse_func;

    __debug '__getfile ' . $filename;

    my $parse = 0;
    my $ret = undef;

    $filename = __effective_filename($filename);

    # Pour ne pas passer son temps  faire des __stat tout le temps, on ne le fait
    # que toutes les minutes pour un mme fichier dj en mmoire.
    my $now = time;
    if (defined $files{$filename}{'data'}) {
	if ($files{$filename}{'dt_last_check'}) {
	    # __debug "  dt_last_check = " . $files{$filename}{'dt_last_check'} . " et now = " . $now;
	    if ($now - $files{$filename}{'dt_last_check'} < 60) {
		__debug "  Fichier $filename dj en mmoire et je ne teste pas cette fois-ci s'il n'est pas obsolte";
		return $files{$filename}{'data'};
	    }
	    else {
		__debug "  Fichier $filename dj en mmoire, je vais tester s'il n'est pas obsolte";
	    }
	}
	else {
	    error "  Fichier $filename dj en mmoire MAIS pas de dt_last_check ! Je recharge tout.";
	    delete $files{$filename};
	}
    }

    my $binfile = $filename . '.bin';

    my $st_file = __stat $filename;
    unless ($st_file) {
	error "  __stat $filename: $!";
	return undef;
    }

    # Si je suis sur un fichier avec des conditions, je saute
    # le retrieve du .bin
    return __parse_file($filename, undef, $parse_func)
	if ($files{$filename}{condition_variable}); 
    

    my $st_bin = __stat $binfile;

    if ($st_bin) {
	if ($st_file->mtime >= $st_bin->mtime) {
	    __debug "  parse $filename ($binfile obsolte)";
	    $parse = 1;
	}
	elsif (    not defined $files{$filename}
		or not defined $files{$filename}{'dtin'}
		or not defined $files{$filename}{'data'}
		or $st_bin->mtime > $files{$filename}{'dtin'} # Gruik : surtout pas >= ici contrairement  au-dessus
	    ) {
	    __debug "  retrieve $binfile ...";
	    eval { $ret = retrieve $binfile; };
	    if ($@) {
		chomp $@;
		error "  retrieve: $@";
		$parse = 1;
	    }
	    else {
		__debug "  retrieve $binfile OK, dtin=" . $st_bin->mtime;
		$files{$filename}{'data'} = $ret;
		$files{$filename}{'dtin'} = $st_bin->mtime;
		$files{$filename}{'dt_last_check'} = $now;
	    }
	}
	else {
	    __debug "  pas besoin de retrieve $binfile: deja en memoire et pas obsolete";
	    $files{$filename}{'dt_last_check'} = $now;
	    $ret = $files{$filename}{'data'};
	}
    } else {
	__debug "  __stat $binfile: $!";
	$parse = 1;
    }

    # Doit-on reparser les eventuelles "dependances" ?
    if ( !$parse ) {
	if ( defined ( $ret->{'  depends'} ) && ref($ret->{'  depends'}) eq 'ARRAY') {
	    my $tmp = {};
	    for my $dep ( @{$ret->{'  depends'}} ) {
		next if $tmp->{$dep};
		$tmp->{$dep} = 1;
		$dep = __effective_filename($dep);
		if ( ! -r $dep ) {
		    error "File not found $dep, required from $filename as a dependency, consider unmodified.";
		    next;
		}
		my $st_dep = __stat $dep;
		if ( ! defined $st_dep ) {
		    error "Cannot stat $dep, required from $filename as a dependency, consider unmodified.";
		    next;
		}
		if ( $st_dep->mtime >= $st_bin->mtime ) {
		    __debug "  parse $filename ($dep obsolte au moins), on ne controle pas les autres";
		    $parse=1;
		    last;
		}
	    }
	}
    }

    if ($parse) {
	__debug "  au final, il faut reparser $filename";
	$ret = __parse_file($filename, $binfile, $parse_func);
	if (defined $ret) {
	    $files{$filename}{'data'} = $ret;
	    $files{$filename}{'dtin'} = $files{$filename}{'dt_last_check'} = time;
	    __debug "  et dtin = " . $files{$filename}{'dtin'};
	    __debug "  et dt_last_check = " . $files{$filename}{'dt_last_check'};
	}
	else {
	    error "parse_file a echoue !!";
	}
    }

    return $ret;
}


sub __add_new_section($$) {
	my ( $res, $section ) = @_;
	$res->{$section} = {} if ! exists $res->{$section};
}

sub __add_new_val($$$$) {
	my ( $res, $section, $key, $val ) = @_;
	if ( ! defined $section ) {
		error "Found $key while not yet in a section. Reject.";
		return;
	}
	if ( defined $res->{"\tencoding"} && $res->{"\tencoding"} eq 'utf8' ) {
		Encode::_utf8_on($val);
		Encode::_utf8_on($key);
	}
	if ( $key =~ /^@/ ) {
		if ( ! defined $res->{$section}->{$key} ) {
			$res->{$section}->{$key} = [ $val ];
		} else {
			push @{$res->{$section}->{$key}}, $val;
		}
	} else {
		$res->{$section}->{$key} = $val;
	}
}

sub __unroll_loop_section($$$$$;$$);
sub __unroll_loop_section($$$$$;$$) {
	my ( $res, $section, $name, $vars, $vals, $loop, $cnt ) = @_;
	$loop //= [ @$vars ];
	$cnt  //= {};
	if ( @$loop ) {
		my $var = shift @$loop;
		for my $val ( @{$vals->{$var}} ) {
			$cnt->{$var} = $val;
			__unroll_loop_section($res, $section, $name, $vars, $vals, $loop, $cnt);
		}
		unshift @$loop, $var;
		return;
	}
	my $item = [];
	my $sec_name = $name;
	for my $var ( @$vars ) {
		my $val = $cnt->{$var};
		$sec_name =~ s/@\($var\)/$val/;
		push @$item, $cnt->{$var};
	}
	__add_new_section($res, $sec_name);
	for my $key ( keys %$section ) {
		my $val;
		my $tmp_items = [ @$item ];
		while ( @$tmp_items ) {
			my $tmp = $key.'('.join(',',@$tmp_items).')';
			if ( exists $res->{$name}->{$tmp} ) {
				$val = $res->{$name}->{$tmp};
				last;
			}
			pop @$tmp_items;
		}
		if ( ! defined $val ) {
			$val = $section->{$key};
		}
		for my $var ( @$vars ) {
			$val =~ s/@\($var\)/$cnt->{$var}/gs;
		}
		__add_new_val($res,$sec_name,$key,$val);
	}
}

sub __unroll_section($$) {
	my ( $res, $name ) = @_;
	my $vars = [];
	my $vals = {};
	my $tmp = $name;
	my $invalid = 0;
	my $section = { %{$res->{$name}} };
	while ( $tmp =~ s/@\(([a-zA-Z0-9]*)\)/x/ ) {
		my $var = $1;
		if ( ! exists $section->{$var} ) {
			error "In loop-section $name, variable '$var' is not defined.";
			$invalid = 1;
			next;
		}
		my $val = delete $section->{$var};
		if ( $val !~ /^(\S+\s*,\s*)+\S+$/ ) {
			error "In loop-section $name, variable '$var' has an invalid definition.";
			$invalid = 1;
			next;
		}
		$vals->{$var} = [ split /\s*,\s*/, $val ];
		push @$vars, $var;
	}
	my $nb_args = scalar(@$vars)-1;
	for my $key ( keys %$section ) {
		next if $key !~ /\(.*\)$/;
		my $val = delete $section->{$key};
		if ( $key !~ /\((\S+,){0,$nb_args}\S+\)/ ) {
			error "Invalid number of arguments in '$key', in loop-section '$name'.";
			$invalid = 1;
		}
	}
	if ( $invalid ) {
		error "File is not valid.";
		return;
	}
	__unroll_loop_section($res, $section, $name, $vars, $vals);
	delete $res->{$name};
}

sub __parse_ini($) {
	my ( $inifile ) = @_;

	__debug '__parse_ini'.$inifile;
	return undef unless defined $inifile;
	my $res = {};

	if ( ! open CONFIG, "<$inifile" ) {
		error "    Le fichier $inifile n'existe pas ou ne peut pas tre lu";
		return $res;
	}

	my ($section, $key, $val, $next);
	my $line;
	while ($line = <CONFIG>) {
		chomp $line;
		if ( $line =~ /^\s*#\s+file charset = (?:utf8|utf-8)\s*/ ) {
			$res->{"\tencoding"} = 'utf8';
		}
		$line =~ s/\s*$//;
		next if $line =~ /^\s*#.*$/;

		if ( $line =~ /^\[([^\]]+)\]\s*$/ ) {
			$section = $1;
			if ( defined $res->{"\tencoding"} && $res->{"\tencoding"} eq 'utf8' ) {
				Encode::_utf8_on($section);
			}
			__add_new_section($res,$section);
			next;
		}
		if ( $line =~ /^\s*(\S*?)\s*=\s*<<(.*)$/ ) {
			$key = $1;
			$val = '';
			my $delim = $2;
			while ( $next = <CONFIG> ) {
				last if ( $next eq $delim );
				$val .= $next;
			}
			__add_new_val($res,$section,$key,$val);
			next;
		}
		if ( $line =~ /^\s*(\S*?)\s*=\s*(.*)$/ ) {
			$key = $1;
			$val = $2;
			while ( $val =~ s/\\$// ) {
				$next = <CONFIG>;
				last if ! defined $next;
				$next =~ s/\s*(?:#.*)?$//;
				$val .= $next;
			}
			__add_new_val($res,$section,$key,$val);
		}
	}
	close CONFIG;

	# Le hash est complet. Resoudre les references (plus generalement les fonctions)
	foreach $section (keys %$res) {
		next if $section =~ /^\s/;
		foreach $key (keys %{$res->{$section}}) {
			__debug '__parse_ini: cherche code dans '.$section.' / '.$key;
			if ( $key =~ /^@/ ) {
				__expand_key_array ($res, $section, $key);
			} else {
				__expand_key_scalar($res, $section, $key);
			}
		}
	}
	foreach $section ( keys %$res ) {
		next if $section =~ /^\s/;
		next if $section !~  /@\([a-zA-Z0-9]+\)/;
		__unroll_section($res, $section);
	}
	return $res;
}

sub __expand_key_array($$$) {
	my ( $res, $section, $key ) = @_;
	my $retry=1;
	while ($retry) {
		my @res=();
		$retry=0;
		for my $val (@{$res->{$section}->{$key}}) {
			if ( $val =~ /^%[a-x]/) {
				$val = __ini_expand($res,$val);
				$retry=1;
			}
			if ( ref($val) eq 'ARRAY' ) {
				__debug "While expanding $section/$key, found an ARRAY-ref";
				push @res,@$val;
			} else {
				__debug "While expanding $section/$key, found a scalar";
				push @res,$val;
			}
		}
		$res->{$section}->{$key}=\@res;
	}
}

sub __expand_key_scalar($$$) {
	my ( $res, $section, $key ) = @_;
	my $val = $res->{$section}->{$key};
	while ( defined $val && $val =~ /^%[a-x]/ ) {
		$val = __ini_expand($res,$val);
		$res->{$section}->{$key} = $val;
	}
}


# Pour toute valeur de clef non simple, cette fonction va en faire
# l'expansion. On lui passe le hash correspondant a l'etat actuel
# de la config qu'on est en train de lire, et le texte de la
# clef a expander.
#
# On commence par traduire la valeur de cette clef en une liste de
# token. Chaque token peut etre de type:
# - fct: appel de fonction, la valeur est le nom (%ref, %cat, etc)
# - open: ouverture de parenthese, la valeur est une parenthese ouvrante
# - close: fermeture de parenthese, la valeur est une parenthese fermante
# - text: texte libre
# - err: une erreur.
#
# Pour decouper en token, on fait un split convenablement choisi sur
# - nom de fonction
# - parenthese ouvrante non precedee d'un \
# - parenthese fermante non precedee d'un \
#
# Ensuite, on type les token, puis on analyse le resultat.
#
# C'est en toute fin d'expansion qu'on supprime les protections sur les
# virgules et les parentheses
sub __ini_expand($$) {
    my ( $ini, $val ) = @_;
    return $val if ( ref($val) );
    my $tokens = __ini_tokenize($val);
    # __debug Dumper($tokens);
    
    __ini_analyse($ini,$tokens);
    
    if ( $tokens->[0]->{'type'} eq 'err' ) {
	warning "Erreur: ".$tokens->[0]->{'val'};
    }
    my $res = $tokens->[0]->{'val'};
    
    $res = '' if ! defined $res;
    $res =~ s/\\(,|\(|\))/$1/g;
    __debug "Resultat de l'expansion: $res";
    
    return $res;
}

sub __ini_tokenize($) {
    my ( $text ) = @_;
    my @tokens = split /(%[a-z]*|(?<!\\)\(|(?<!\\)\))/, $text;
    @tokens = grep(!/^$/,@tokens);
    map { 
	my $txt = $_;
	$_ = { 'val' => $txt };
	if ( $txt =~ /^%[a-z]*$/ ) {
	    $_->{'type'} = 'fct';
	} elsif ( $txt eq '(' ) {
	    $_->{'type'} = 'open';
	} elsif ( $txt eq ')' ) {
	    $_->{'type'} = 'close';
	} else {
	    $_->{'type'} = 'text';
	}
    } @tokens;
    return \@tokens;
}

# Re-place un token en tete de liste
sub __tok_unshift($$$) {
    my ( $toks, $type, $val ) = @_;
    unshift @$toks, {'type'=>$type,'val'=>$val};
}

# La fonction __ini_analyse implemente un automate a pile. Le principe
# est simple. Elle prend en parametres le hash correspondant a la conf
# en cours d'analyse (premier parametre, pour permettre les references
# internes) et la liste des tokens correspondant a la commande en cours
# d'analyse.
#
# Quelle que soit la liste (valide) de tokens qu'on lui donne, apres execution,
# le promier token de la liste est:
# - soit un token de type err, si une erreur s'est produite a un moment
#   dans l'expansion
# - soit un token de type text, qui est le resultat de l'expansion de la
#   commande qu'on a trouve.
#
# Cette fonction est massivement recursive.
# - si on lui donne une liste de token qui commence par un texte ou une
#   erreur, il n'y a plus rien a faire (soit tout est perdu, soit on est
#   arrive en bout de chaine, sur un element qui n'a pas besoin d'etre
#   analyse)
# - si on lui donne une liste de token qui commence par un open (parenthese
#   ouvrante), elle consomme tous les tokens jusqu'a la parenthese fermante inclue
#   et les concatene en un token texte (apres les avoir analyses, bien entendu).
#   C'est ici qu'a lieu la recurrence.
#   Le token texte resultant est place en debut de liste.
# - si on lui donne une liste de token qui commance par une fonction, on
#   execute la fonction, et on place le token resultant de l'execution
#   (un texte ou une erreur) en tete de liste.
# - si on lui donne une liste de token qui commence par un close, alors
#   c'est une erreur: la regle de traitement des opens se charge de consommer
#   les closes correspondant. S'il en reste un, c'est qu'il n'a pas open qui
#   lui corresponde.
# - si la liste est vide, c'est une erreur.
sub __ini_analyse($$);
sub __ini_analyse($$) {
    my ( $ini, $toks ) = @_;
    __debug 'Entree dans __ini_analyse';
    # Conditions d'arret:
    if ( ! scalar @$toks ) {
	__tok_unshift($toks,'err','Expression mal terminee, sans doute une parenthese qui manque');
	return;
    }
    if ( $toks->[0]->{'type'} eq 'text' ) {
	return;
    }
    if ( $toks->[0]->{'type'} eq 'err' ) {
	return;
    }
    my $first = shift @$toks;
    use Data::Dumper;
    __debug 'Analyse du token non-simple: '.Dumper($first);
    __debug 'Analyse du token non-simple: '. $first;
    
    # Si c'est une fonction, elle doit etre suivie d'un argument
    # qui commence par un open. Si l'argument a ete convenablement
    # lu, alors, on execute. Si le resultat est bon, on le retourne.
    # Dans tous les autres cas, c'est une erreur.
    if ( $first->{'type'} eq 'fct' ) {
	__debug 'Appel de fonction';
	my $fct = $first->{'val'};
	
	if ( $toks->[0]->{'type'} ne 'open' ) {
	    # ERR
	    # die "Fonction sans argument";
	    __tok_unshift($toks,'err',"Fonction sans argument");
	    return;
	}

	# Ici, on analyse en commencant par l'open qui suit la fonction.
	# Le resultat de cette analyse doit, par definition, etre un texte
	# ou une erreur.
	__ini_analyse($ini,$toks);
	my $next = shift @$toks;
	if ( $next->{'type'} eq 'err' ) {
		__tok_unshift($toks,'err','Echec lecture arguments de '.$fct.': '.$next->{'val'});
		return;
	}
	my $args = $next->{'val'};
	__debug 'Fonction '.$fct.' appelee sur '.$args;
	
	# Le traitement nous retourne un token. Soit un texte, soit une erreur.
	# C'est le resultat du traitement, on le place en tete de liste.
	my $res = __ini_exec($ini,$fct,$args);
	unshift @$toks,$res;
	return;
    }
    
    # Si c'est un open, on traite tous les tokens suivants, et on les accumule,
    # jusqu'a croiser soit un close, soit une erreur.
    # Si on croise une erreur, on la remonte telle qu'elle, si on croise un close,
    # on le jette et on retourne ce qu'on a accumule.
    if ( $first->{'type'} eq 'open' ) {
	__debug 'Accumulation argument';
	my $txt;
	while ( $toks->[0]->{'type'} ne 'close' ) {
	    __ini_analyse($ini,$toks);
	    if ( $toks->[0]->{'type'} eq 'err' ) {
		return;
	    }
	    my $tmp = shift @$toks;
	    __debug 'Dans l accumulation j ajoute '.$tmp->{'val'};
	    $txt .= $tmp->{'val'};
	}
	shift @$toks; # Mange le close
	__tok_unshift($toks,'text',$txt);
	return;
    }
    
    # Si c'est un close, c'est qu'on a pas vu l'open correspondant
    # auparavant (le traitement de l'open mange le close associe avant de
    # produire un texte).
    if ( $first->{'type'} eq 'close' ) {
	# ERR
	# die "Parenthese surnumeraire.";
	__tok_unshift($toks,'err',"Parenthese surnumeraire.");
	return;
    }
}

sub __ini_exec($$$) {
	my ( $ini, $fct, $args ) = @_;
	my @args = split /(?<!\\),/,$args;
	my $ret;
	my $type='text';
	if ( $fct eq '%ref' ) {
		my $tmp = $ini->{$args[0]}->{$args[1]};
		$ret=__ini_expand($ini,$tmp);
###		# En Woody, on a beau faire, on rcupre toujours une chane de
###		# caractres multi-lignes au lieu d'une rfrence de liste :(
###		if ($args[1] =~ m/^@/ and ref($ret) ne 'ARRAY') {
###			my @tmp = split("\n", $ret);
###			$ret = \@tmp;
###		}
###		# En Lenny, on a beau faire, on rcupre toujours une chaine de
###		# caractres dans un tableau...
###		if ( $args[1] =~ m/^@/ and ref($ret) eq 'ARRAY' and $#{$ret} == 0 and $ret->[0] =~ m/\n/ ) {
###			my @tmp = split("\n", $ret->[0]);
###			$ret = \@tmp;
###		}
	} elsif ( $fct eq '%refext' ) {
		$ret=Config_Key($args[0],$args[1],$args[2]);
		if (! ref $ini->{'  depends'}) {
			$ini->{'  depends'} = [];
		}
		push @{$ini->{'  depends'}},$args[0];
	} elsif ( $fct eq '%file' ) {
		my $file = __effective_filename($args[0]);
		$ret = slurp_file($file);
		$ini->{'  depends'} //= [];
		push @{$ini->{'  depends'}},$file;
	} elsif ( $fct eq '%cat' ) {
		$ret=join ("",@args);
	} elsif ( $fct eq '%protect') {
		$args =~ s/(?<!\\)(,|\(|\))/\\$1/g;
		$ret=$args;
	} else {
		$type='err';
		$ret='Echec de '.$fct.'('.$args.'): Fonction inconnue';
	}
	return {'type'=>$type,'val'=>$ret};
}


sub __parse_tbl($) {
	my ($tblfile) = @_;
	my %tbl;

	return undef unless defined $tblfile;
	
	__debug '__parse_tbl ' . $tblfile;

	if (! open(CONFIG, "<$tblfile")) {
		__debug "    Le fichier $tblfile n'existe pas ou ne peut pas tre lu";
		return {};
	}

	my ($section, $last_key) = ("");
	while ( my $line = <CONFIG> ) {
		chomp $line;
		#__debug "ligne : $_\n";
		if ( $line =~ /^\s*#\s+file charset = (?:utf8|utf-8)\s*/ ) {
			$tbl{"\tencoding"} = 'utf8';
		}
		$line =~ s/\s*(?:#.*)?$//;
		next if $line =~ m/^$/;
		if ( defined $tbl{"\tencoding"} && $tbl{"\tencoding"} eq 'utf8' ) {
			Encode::_utf8_on($line);
		}
		if ($line =~ m/^\s*{\s*([^:]+)\s*:\s*([^}]+)\s*}\s*$/) {
			__debug "    dfinition de la section $1";
			$section = $1;
			$tbl{$section}{'title'} = $2;
			next;
		}
		if ($section) {
			while ( $line =~ s/([^\\])\\$/$1/ ) {
				my $more = <CONFIG>;
				chomp $more;
				if ( defined $tbl{"\tencoding"} && $tbl{"\tencoding"} eq 'utf8' ) {
					Encode::_utf8_on($more);
				}
				$line .= $more;
			}
			if ($line =~m/^\s*\[\s*([^:]+)?\s*:\s*(.+)\s*\]\s*$/) {
				$last_key = push(@{$tbl{$section}{'cles'}}, { nom => $2, cle => $1 }) - 1;
			} elsif (defined $last_key and $line =~ m/^\s*actions\s*=\s*(.+)\s*\s*$/) {
				my $array;
				@$array = split(/\s*,\s*/, $1);
				$tbl{$section}{'cles'}[$last_key]->{'actions'} = $array;
			} elsif (defined $last_key and $line =~ m/^\s*((?:regexp|code|texte|divclass|spanclass|regwrap|linkto|subst)(?:-[^\s]*)?)\s*=\s*(.*)\s*$/) {
				$2 = "" unless defined $2;
				$tbl{$section}{'cles'}[$last_key]{$1} = $2;
			} elsif (defined $last_key and $line =~ m/^\s*(enum-null(?:-[^\s]*)?)\s*=\s*(.*)\s*$/) {
				$2 = "" unless defined $2;
				$tbl{$section}{'cles'}[$last_key]{$1} = $2;
			} elsif (defined $last_key and $line =~ m/^\s*(enum-defaul?t(?:-[^\s]*)?)\s*=\s*(.*)\s*$/) {
				$2 = "" unless defined $2;
				$tbl{$section}{'cles'}[$last_key]{'enum-default'} = $2;
			} elsif (defined $last_key and $line =~ m/^\s*(enum(?:-[^\s]*)?)\s*=\s*(.*?)\s*=>\s*(.*)\s*$/) {
				## La cl peut tre vide !
				$tbl{$section}{'cles'}[$last_key]{"$1"}{$2} = $3;
			}
		}
	}
	close CONFIG;
	return \%tbl;
}


sub __condition_parse_test ($$) {
    my ($vars, $var) = @_;

    foreach (@$vars) {
	return 1 if $_ eq $var;
    } 

    return 0;
}


sub __condition_parse ($$$) {
    my ($file_source, $file_cible, $vars) = @_;

    my @condission;
    open SOURCE,"<".$file_source; # FIXME error ?
    open CIBLE, ">$file_cible" or error "Can't open file_cible file : $file_cible";

    my @if;
    foreach my $line (<SOURCE>) {
	my $jump;
	next if ($line =~ /^\s*$|^\s*\#/);
	my $buffer;

	if ( @condission) {
	    $jump = "Jump!";
	    my $condi = $condission[@condission -1];
	    
	    if ( __condition_parse_test ($vars, $condi->{val} ) and
		($condi->{etat} == 1)
		) {
		$buffer = $line;
	    }
	    else {
		if ((! __condition_parse_test ($vars, $condi->{val} )) and
		    ($condi->{etat} == 0) and
		    ($condi ne $if[@if -1])
		    ) {
		    $buffer = $line; 
		}
		elsif ((! __condition_parse_test ($vars, $condi->{val} )) and
		    ($condi->{val} eq $if[@if -1]) and
		    ($condi->{etat} == 0) and
		    (__condition_parse_test($vars,$if[@if -1]))
		    ) {
		    undef $jump;
		}
	    }
	}

	# test des conditions
	if ($line =~ /^\@if\s+(\S+)$/) {
	    my $if = $1;
	    push @if, $if;
	    push @condission, { val => $if, etat => 1 };
	    next;
	}
	elsif ($line =~ /^\@else$/) {
	    die "Pas de if avant le else ! Ca ne me plait pas, mais alors pas du tout...\n"
		unless (@condission);

	    $condission[@condission -1]->{etat} = 0;
	    next;
	}
	elsif ($line =~ /^\@endif$/) {
	    pop @if;
	    pop @condission;
	    next;
	}

	# Impression dans fichier tmp
	if ($buffer) {
	    print CIBLE $buffer;
	    next;
	}
	print CIBLE $line
	    unless $jump;
    }

    close SOURCE or error "Can't close SOURCE\n";
    close CIBLE or error "Can't close CIBLE\n";
}


##############################################################################
#
#   Fonctions exportes
#

=head3 Config_Files_List($dir)

Return config file list from directory $dir in alphabetic order.

=cut
sub Config_Files_List($) {
	my ( $config_dir ) = @_;

	if ( ! -d $config_dir ) {
		error("Can't read $config_dir. It's not a directory.");
		return undef;
	}
	if ( ! -x $config_dir ) {
		error("Can't read $config_dir. It's not executable.");
		return undef;
	}

	my @config_files;
	my $dh;
	opendir $dh, $config_dir;
	while ( my $fname = readdir $dh ) {
		next if $fname =~ /^\./;
		next if $fname !~ /\.cfg$/;
		push @config_files, "$config_dir/$fname";
	}
	closedir $dh;

	@config_files = sort { $a cmp $b } @config_files;
	return \@config_files;
}

sub Unload_Config ($) {
    my ($file) = @_;

    return undef
	unless defined $file;

    my $filename = __effective_filename $file;
    __debug 'Unload_Config ' . $filename;
    delete $files{$filename};

    return 1;
}


sub Forget_Config($) {
    my $filename = shift;

    return undef unless defined $filename;
    $files{__effective_filename($filename)} = undef;
}



sub Load_Config($) {
    my ($file) = @_;

    return undef
	unless defined $file;

    __debug 'Load_Config ' . $file;
    return __get_file $file, \&__parse_ini;
}


sub Config_Sections_List($) {
    my ($file) = @_;

    return undef
	unless defined $file;

    my $ref = __get_file $file,\&__parse_ini;

    return undef
	unless defined $ref;

    return undef
	unless ref $ref eq 'HASH';

    my @list = grep !/^  depends$/,keys %$ref;
    return \@list;
}


sub Config_Section($$) {
    my ($file, $section) = @_;

    return undef
	if ( $section eq '  depends' );
    return undef
	unless defined $file and defined $section;

    my $ref = __get_file $file,\&__parse_ini;

    return undef
	unless defined $ref;

    return undef
	unless defined $ref->{$section};

    return $ref->{$section};
}


sub Config_Section_Keys($$) {
    my ($file, $section) = @_;

    return undef
	if ( $section eq '  depends' );
    return undef
	unless defined $file and defined $section;

    my $ref = Config_Section $file, $section;

    return undef
	unless defined $ref;

    return undef
	unless ref $ref eq 'HASH';

    my @list = keys %$ref;
    return \@list;
}


sub Config_Key($$$) {
    my ($file, $section, $key) = @_;

    return undef
	unless defined $file and defined $section and defined $key;
    return undef
	if ( $section eq '  depends' );

    my $ref = Config_Section $file, $section;

    return undef
	unless defined $ref;

    return undef
	unless ref $ref eq 'HASH';

    return undef
	unless defined $ref->{$key};

    # En Woody, on a beau faire, on rcupre toujours une chane de
    # caractres multi-lignes au lieu d'une rfrence de liste :(
    if ($key =~ m/^@/ and ref($ref->{$key}) ne 'ARRAY') {
	my @tmp = split("\n", $ref->{$key});
	return \@tmp;
    }
    # En Lenny, on a beau faire, on rcupre toujours une chaine de
    # caractres dans un tableau...
    if ( $key =~ /^@/ and ref($ref->{$key}) eq 'ARRAY' and $#{$ref->{$key}} == 0 and $ref->{$key}->[0] =~ m/\n/ ) {
	my @tmp = split("\n", $ref->{$key}->[0]);
	return \@tmp;
    }

    return $ref->{$key};
}



sub Rewrite_Entry($$$$) {
	my ( $clef, $val, $values, $rules ) = @_;

	my $res = $val;
	my $show_res = $val;
	$show_res = '<UNDEF>' if ! defined $val;
	$clef = '<UNDEF>' if ! defined $clef;
	__debug "Before rewrite, $clef is $show_res.";
	if ( ! exists $rules->{'actions'} or
	     ref ($rules->{'actions'}) ne 'ARRAY' ) {
		__debug "No rewrite rule. Returning.";
		return $res;
	}
	for my $act ( @{$rules->{'actions'}} ) {
		next if ( ! defined $rules->{$act} );
		
		if ( $act =~ /^texte/ ) {
			$res = $rules->{$act};
			goto NEXT;
		}
		if ( $act =~ /^divclass/ ) {
			my $cl = $rules->{$act};
			$res = '' if ( ! defined $res );
			$res = '<div class="'.$cl.'">'.$res.'</div>';
			goto NEXT;
		}
		if ( $act =~ /^spanclass/ ) {
			my $cl = $rules->{$act};
			$res = '' if ( ! defined $res );
			$res = '<span class="'.$cl.'">'.$res.'</span>';
			goto NEXT;
		}
		
		if ( $act =~ /^regexp/ ) {
			if ( defined $res ) {
				eval('$res =~ '.$rules->{$act});
			}
			goto NEXT;
		}

		if ( $act =~ /^regwrap/ ) {
			if ( defined $res ) {
				my $tmp = $rules->{$act};
				$tmp =~ s/\//\\\//g;
				my $re = 's/^(.*)$/'.$tmp.'/gs';
				eval('$res =~ '.$re);
			}
			goto NEXT;
		}

		if ( $act =~ /^linkto/ ) {
			if ( defined $res ) {
				my $tmp = $rules->{$act};
				$tmp =~ s/\//\\\//g;
				my $re = 's/^(.*)$/<a href="'.$tmp.'">$1<\/a>/g';
				eval('$res =~ '.$re);
			}
			goto NEXT;
		}

		my $keys = join('|', keys %$values);

		if ( $act =~ /^subst/ ) {
			my $tmp = $rules->{$act};
			$tmp =~ s/($keys)/$values->{$1}/ge;
			$res =~ s/CURRENT/$res/g;
			goto NEXT;
		}

		my $translate = sub($) {
			my $cle = shift;
			my $v = $values->{$cle} // "";
			$v =~ s/'/\\'/g;
			return $v;
		};
		if ( $act =~ /^code/ ) {
			my $code = $rules->{$act};
			$code =~ s/($keys)/$translate->($1)/ge;
			$res = eval ( $code );
			$res = $@ if $@;
			goto NEXT;
		}
		
		if ( $act =~ /^enum(.*)$/ ) {
			my $suffix = $1;
			if ( ! defined $res ) {
				if ( exists $rules->{'enum-null'.$suffix} ) {
					$res = $rules->{'enum-null'.$suffix};
				} else {
					$res = $rules->{'enum-default'.$suffix};
				}
				next;
			}
			if ( defined $rules->{$act}->{$res} ) {
				$res = $rules->{$act}->{$res};
			} else {
				$res = $rules->{'enum-default'.$suffix};
			}
		}
NEXT:
		$show_res = $res;
		$show_res = '<UNDEF>' if ( ! defined $show_res );
		__debug "After rule $act we have $show_res.";
	}
	return $res;
}

sub Rewrite_Props($$$$;$);
sub Rewrite_Props($$$$;$) {
	my ( $file, $table, $template, $values, $titre ) = @_;
	if ( ! defined $file or
	     ! defined $table or
	     ! defined $values ) {
		return undef;
	}
	
	__debug "Rewrite properties with table $table.";
	my %done;
	my @return;
	my $tbl = __get_file $file, \&__parse_tbl;
	if ( ! defined $tbl ) {
		goto otherkeys;
	}

	if ( ! defined $tbl->{$table} ) {
		goto otherkeys;
	}
	$tbl = $tbl->{$table};
	$titre = $tbl->{'title'} if ! defined $titre;
	
	if ( ! defined $tbl->{'cles'} ) {
		goto otherkeys;
	}
	$tbl = $tbl->{'cles'};

	for my $rules ( @$tbl ) {
		my $key = $rules->{'cle'};
		next if ( ! defined $key );
#		next if ( ! defined $values->{$key} );
		my $val = $values->{$key};
		
		if ( ref($val) eq 'HASH' ) {
			__debug "$key is an HASH, switching to sub-table.";
			$val = Rewrite_Props($file, $table.'/'.$key, $template, $val);
			push @return, {
				'nom' => $rules->{'nom'},
				'valeur' => $val
				};
		} elsif ( ref($val) eq 'ARRAY' ) {
			my $subval;
			__debug "$key is an ARRAY, looping.";
			for my $v ( @$val ) {
				if ( ref($v) eq 'HASH' ) {
					__debug "This time $key is an HASH, switching to sub-table.";
					$subval = Rewrite_Props($file, $table.'/'.$key, $template, $v);
				} elsif ( ref($v) eq 'ARRAY' ) {
					error "I cannot have an array in an array within properties.";
					# C'est impossible dans les proprietes
				} else {
					__debug "This time $key is a scalar. I rewrite it.";
					$subval = Rewrite_Entry($key,$v,$values,$rules);
				}
				push @return, {
					'nom' => $rules->{'nom'},
					'valeur' => $subval
					};
			}
		} else {
			__debug "$key is a scalar. I rewrite it.";
			$val = Rewrite_Entry($key,$val,$values,$rules);
			push @return, {
				'nom' => $rules->{'nom'},
				'valeur' => $val
				};
		}
		$done{$key}=1;
	}
otherkeys:
	__debug "No more information in that table. Dumping remaining keys.";
	for my $key ( keys %$values ) {
		next if ( exists $done{$key} );
		my $val = $values->{$key};
		if ( ref($val) eq 'HASH' ) {
			__debug "$key is an HASH. Switching to sub-table.";
			$val = Rewrite_Props($file, $table.'/'.$key, $template, $val);
			push @return, {
				'nom' => $key,
				'valeur' => $val
				};
		} elsif ( ref($val) eq 'ARRAY' ) {
			__debug "$key is an ARRAY. Looping.";
			my $subval;
			for my $v ( @$val ) {
				$subval = $v;
				if ( ref($v) eq 'HASH' ) {
					__debug "This time $key is an HASH. Switching to sub-table.";
					$subval = Rewrite_Props($file, $table.'/'.$key, $template, $v);
				} elsif ( ref($v) eq 'ARRAY' ) {
					error "I should not have an array in an array for properties.";
				} else {
					__debug "This time $key is a scalar. No rewrite since no more rules.";
				}
				push @return, {
					'nom' => $key,
					'valeur' => $subval
					};
			}
		} else {
			__debug "$key is a scalar. No rewrite since no more rules.";
			push @return, {
				'nom' => $key,
				'valeur' => $values->{$key}
				};
		}
	}
	my $res = { titre => $titre, valeurs => \@return };
	conduse('FDN::Web::TemplateParse');
	return remplace($template,$res);
}

sub Rewrite_Table($$$) {
	my ($file, $table, $values) = @_;
	if ( ! defined $file or
	     ! defined $table or
	     ! defined $values ) {
		return undef;
	}
	
	my $tbl = __get_file $file, \&__parse_tbl;
	if ( ! defined $tbl ) {
		return undef;
	}

	if ( ! defined $tbl->{$table} ) {
		return undef;
	}
	my $encoding = 'text';
	if ( exists $tbl->{"\tencoding"} ) {
		$encoding = $tbl->{"\tencoding"};
	}
	$tbl = $tbl->{$table};
	my $titre = $tbl->{'title'};

	if ( ! defined $tbl->{'cles'} ) {
		return undef;
	}
	$tbl = $tbl->{'cles'};

	my @return;
	for my $rules (@$tbl) {
		# __debug 'Dbut foreach, $hash = ' . Dumper($hash);
		my $key = $rules->{'cle'};
		next if ( ! defined $key );
		my $val;
		$val = $values->{$key} if ( defined $values->{$key} );
		$val = Rewrite_Entry($key,$val,$values,$rules);
		push @return, {
			'nom' => $rules->{'nom'},
			'valeur' => $val
			};
	}

	return { titre => $titre, valeurs => \@return, encoding => $encoding };
}


# renvoie une string representant la config
# parametre: hash de config
# retour   : la string si tout s'est bien passe, undef sinon
sub Create_Config ($) {
    my ($struct) = @_;

    unless (defined $struct and ref($struct) eq 'HASH') {
	error "Paramtre 'struct' non dfini ou pas un HREF";
	return undef;
    }

    my $conf_string = '';
    foreach my $section (keys %{$struct}) {
	if (ref($struct->{$section}) ne 'HASH') {
	    error "Section '$section': pas un HREF";
	    return undef;
	}

	$conf_string .= "[$section]\n";
	foreach my $item (keys %{$struct->{$section}}) {
	    if ($item =~ /^@/) {
		if (ref($struct->{$section}->{$item}) ne 'ARRAY') {
		    error "Section '$section' Item '$item': pas un AREF";
		    return undef;
		}

		foreach my $element (@{$struct->{$section}->{$item}}) {
		    $conf_string .= "$item = $element\n";
		}
	    }
	    else {
		if (ref($struct->{$section}->{$item})) {
		    error "Section '$section' Item '$item': pas un scalaire";
		    return undef;
		}
		$conf_string .=  "$item = $struct->{$section}->{$item}\n";
	    }
	}
	$conf_string .= "\n";
    }

    return $conf_string;
}


# ecrit un fichier de conf 
# parametre	: nom du fichier de conf
#		: hash de config
# retour	: 0 si tout c'est bien passe
sub Write_Config ($$) {
    my ($file_t, $struct) = @_;
	# untaint le nom du fichier pour pouvoir ecrire des 
	# fichier de conf depuis un script setuid
	$file_t =~ /(^.*$)/;
	my $file = $1;


    unless (defined $file and length $file) {
	error "Paramtre 'file' non dfini ou vide";
	return 1;
    }

    unless (defined $struct) {
	error "Paramtre 'struct' non dfini";
	return 1;
    }

    my $config = Create_Config $struct;
    if (not defined $config) {
	error "Create_config failed";
	return 1;
    }
    else {
	# FIXME ne faudrait-il pas s'assurer que le fichier gnr
	# est en criture pour le groupe sitacfg ? Voir la gnration du .bin
	# pour une solution possible.
	my $effective_filename = __effective_filename($file);
	unless (open HANDLER, ">$effective_filename") {
	    error "open >$effective_filename failed: $!";
	    return 1;
	}

	print HANDLER $config;
	close HANDLER;
    }
    Forget_Config($file);

    __debug "Write_config($file): OK";
    return 0;
}

# Cette fonction est l pour dfinir des variables conditionnelles qui seront
# utilises lors du parsing conditionnel d'un fichier de configuration.
sub Config_Set_Var ($$) {
    my $filename = __effective_filename (shift);
    my $condition_variable = shift;

    return 0 unless $filename and defined $filename;
    return 0 unless $condition_variable and defined $condition_variable;

    unless (ref($condition_variable)  eq 'ARRAY') {
	error "Le paramtre condition_variable n'est pas un AREF, je l'ignore";
	return 0;
    }

    $files{$filename}{condition_variable} = $condition_variable;
    return 1;
}


##############################################################################
#
#   Fin du package
#
1;

