# This file is part of the "FDN tools" project.
# You can use and redistribute it under the terms of the
# GNU General Public License.
# (c) French Data Network 1997-2005
# Authors and contributors:
# - Benjamin Bayart
# - Nicolas Baradakis
# - Stephane Jourdois
package FDN::Common::FileLock;
#
# Locks sur des filehandles (y compris via NFS)
#
use strict;
use Exporter;
our @ISA = ('Exporter');
our @EXPORT = qw (
	setlock
	unsetlock
	locked
);

use Fcntl;
use POSIX qw(:unistd_h :errno_h);
use FDN::Common::Constantes;

# /!\ OS-dependent structure
# Linux struct flock
#   short l_type;
#   short l_whence;
#   off_t l_start;
#   off_t l_len;
#   pid_t l_pid;
# c2ph says: typedef='s2 l2 i', sizeof=16
# With 32-bits file offset :
# my $FLOCK_STRUCT = 's2l2i';
# With 64-bits file offset :
my $FLOCK_STRUCT = 's2q2i';


# Pose un lock en criture (F_WRLCK) sur le filehandle $fh
# Si $wait est vrai (par dfaut $wait est faux), la fonction ne retourne pas
#   d'erreur si un lock existe dj, mais attend qu'il se libre.
# Si start et len sont >0, ils sont utiliss comme rgion du fichier  locker
#   (cf. man 2 fcntl,  F_GETLK,  F_SETLK  et  F_SETLKW)
# Cette fonction renvoie 1 si la pose du lock a russi, 0 sinon
# Warning : Le fh doit tre ouvert en criture pour pouvoir poser le lock
sub setlock ($;$$$) {
	my ($fh, $wait, $start, $len) = @_;
	$wait = 0 unless defined $wait;
	$start = 0 unless defined $start;
	$len = 0 unless defined $len;

					#type     whence    start   till  pid
	my $packed = pack($FLOCK_STRUCT, F_WRLCK, SEEK_SET, $start, $len, 0);
	my $option;
	if ( $wait ) {
		$option = F_SETLKW;
	} else {
		$option = F_SETLK;
	}
	my $res;
	$res = fcntl($fh, $option, $packed);
	if ($res) {
		return TRUE;
	} else {
		return FALSE;
	}
}

# Supprime le lock sur le filehandle $fh
# Si start et len taient fournis  l'appel de lock, il faut les redonner ici
# Renvoie 1 si le retrait du lock a russi, 0 sinon
#   (sinon = on n'est pas le propritaire du lock)
sub unsetlock ($;$$) {
	my ($fh, $start, $len) = @_;
	$start = 0 unless defined $start;
	$len = 0 unless defined $len;

	my $packed = pack($FLOCK_STRUCT, F_UNLCK, SEEK_SET, $start, $len, 0);
	my $res = fcntl($fh, F_SETLKW, $packed);
	if ($res) {
		return TRUE;
	} else {
		return FALSE;
	}
}

# Renvoie vrai (int > 0) si le $fh a un lock pos sur lui
# Renvoie 1 si ce lock est pos par n'importe-quel process, ou un process inconnu
# Renvoie 2 si le lock est pos par le processus en cours
# Renvoie 0 si aucun lock n'est pos
# Renvoie undef si la fonction choue
# FIXME: Cette fonction ne marche pas. Elle retourne 0 s'il n'y a pas de verrou ou
# si le verrou est detenu par le process en cours (voir man 2 fcntl a ce sujet).
sub locked ($) {
	my ($fh) = @_;

	my $packed = pack($FLOCK_STRUCT, F_WRLCK, SEEK_SET, 0, 0, 0);
	my $res = fcntl($fh, F_GETLK, $packed);
	if (! $res) {
#		print STDERR "Obtenu $res sur fcntl\n";
		return undef;
	} else {
		my ($type, $whence, $start, $len, $pid) = unpack($FLOCK_STRUCT, $packed);
#		print STDERR "Type: $type\nWhence: $whence\nStart: $start\nLen: $len\nPid: $pid\n";
#		print STDERR "WR: ".F_WRLCK."\n";
#		print STDERR "UN: ".F_UNLCK."\n";
		if ($type == F_WRLCK) {
			if ($pid == $$) { return 2 }
			else { return 1 }
		} elsif ($type == F_UNLCK) {
			# Aucun lock n'est positionn
			return 0;
		} else {
			# houla...
			return undef;
		}
		
	}
}

1;
