# This file is part of the "FDN tools" project.
# You can use and redistribute it under the terms of the
# GNU General Public License.
# (c) French Data Network 1997-2005
# Authors and contributors:
# - Benjamin Bayart
# - Stephane Jourdois
# - Alexis Sukrieh
package	FDN::Common::Utils;

use strict;
use Exporter ;

our @ISA = ( 'Exporter' ) ;

our @EXPORT	= qw (
	TIMESTAMP_DB
	TIMESTAMP_14
	TIMESTAMP_8

	get_timestamp
	timestamp14_2_epoch
	epoch_2_timestamp14
	timestamp14_2_db
	timestampdb_2_14
	convert_unit
	convert_time
	make_string_rfc822
	make_alpha_hashing
	make_crc32_hashing
	dir_hashing
	sql14_2_epoch
	epoch_2_sql14
	is_in_array

	mkdir_p
	slurp_file
	slurp_binary
);

use Encode ();

use FDN::Common::Constantes;
use FDN::Common::Modules;
use FDN::Common::Logger;

use constant TIMESTAMP_DB => 'db';
use constant TIMESTAMP_14 => '14';
use constant TIMESTAMP_8  => '8';

sub get_timestamp(;$) {
	# Parametres, variables et constantes locales
	my ( $type ) = @_ ;
	my ( $timestamp ) ;
	conduse('Date::Format');

	if ( ! defined ( $type ) || $type eq TIMESTAMP_DB ) {
		$timestamp = time2str ( "%Y-%m-%d %H\:%M\:%S", time ) ;
	} elsif ( $type eq TIMESTAMP_14 ) {
		$timestamp = time2str ( "%Y%m%d%H%M%S", time ) ;
	} elsif ( $type eq TIMESTAMP_8 ) {
		$timestamp = time2str ( "%Y%m%d", time ) ;
	}
	return $timestamp ;
}

sub timestamp14_2_epoch ($) {
	my ( $date ) = @_;
	$date =~ /^(\d{4})(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)$/;
	my ( $y, $m, $d, $H, $M, $S) = ($1,$2,$3,$4,$5,$6);
	$m--;
	conduse('Time::Local');
	my $temps = timelocal ( $S,$M,$H,$d,$m,$y);
    	return $temps;
}

sub epoch_2_timestamp14 ($) {
	my ($epoch) = @_;
	my ($sec,$min,$hour,$mday,$mon,$year) = localtime($epoch);
	return sprintf("%04d%02d%02d%02d%02d%02d",
		$year + 1900, $mon + 1, $mday, $hour, $min, $sec);
}

sub timestamp14_2_db ($) {
	my ( $date ) = @_;
	$date =~ /^(\d{4})(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)$/;
	return "$1-$2-$3 $4:$5:$6";
}

sub timestampdb_2_14($) {
	my ( $date ) = @_;
	$date =~ /^(\d{4})-(\d\d)-(\d\d)\s+(\d\d):(\d\d):(\d\d)/;
	return "$1$2$3$4$5$6";
}

# Cette fonction prend en entre un nombre entier >= 1, son prfixe (une
# lettre ou rien), un diviseur (en gnral 1000 pour les metres, les litres,
# etc. et 1024 pour les octets, les bits, etc.), et un nombre minimum de
# chiffres significatifs (optionnel, par dfaut 3).
# En contexte de liste, elle renvoit un tableau de deux valeurs :
#  - le nombre donn en entre avec le plus petit nombre possible de chiffres
#      significatifs,
#  - le prefixe correspondant (une lettre)
# En contexte scalaire, elle renvoit une chaine avec le nombre, un espace, et
#   le prfixe.
# En cas d'erreur, la fonction renvoie undef ou une liste vide.
# PS : La fonction ne gre les units pour des nombres >= 1.
# FIXME : valeurs ngatives ?
sub convert_unit ($$$;$) {
	my ($number, $prefix, $divisor, $min_signs) = @_;
	my @prefixs = ('', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y');
	my %prefixs = (
		'' => 0,
		'k' => 1,
		'M' => 2,
		'G' => 3,
		'T' => 4,
		'P' => 5,
		'E' => 6,
		'Z' => 7,
		'Y' => 8
	);
	# (kilo, mega, giga, tra, peta, exa, zetta, yotta)

	$min_signs = 3 unless (defined($min_signs));
	$prefix = '' unless defined $prefix;

	my $multiplier;
	
	unless (defined $prefixs{$prefix} && $divisor>1) {
		return (wantarray ? () : undef);
	}
	$multiplier = $prefixs{$prefix};

	# Tout est l :
	if ($multiplier) {
		$number *= $divisor ** $multiplier;
	}
	$multiplier = 0;
	while ( $number >= $divisor && defined $prefixs[$multiplier+1] ) {
		$number /= $divisor;
		$multiplier++;
	}
	
	# maintenant on supprime les chiffres non significatifs
	# Le nombre de chiffres significatif est le nombre de chiffres avant
	# la virgule + le nombre de chiffres aprs. On ne doit pas supprimer
	# de chiffres avant la virgule, donc il faut garder aprs la virgule
	# autant de chiffres qu'il manque pour avoir le nombre minimum de
	# chiffres significatifs.
	# split renvoie un tableau avec chaque chiffre avant la virgule, et
	# scalar les compte. C'est autant de chiffres significatifs qu'on a
	# dj...
	$number =~ s/(e.*)//;
	my $e=$1 || '';
	my @chiffres = split //, int $number;
	$min_signs -= scalar @chiffres;
	if ($min_signs > 0) {
		$number += .5/(10**$min_signs);
		$number =~ s/^(\d*)((\.\d{0,$min_signs})\d*)?/$1$3/;
		$number = 0+$number;
		$number =~ s/\./,/;
	} else {
		$number =~ s/\..*//;
	}
	$number.=$e;
	# FIXME : ajouter un espace tous les milliers...
	return (wantarray ? ($number, $prefixs[$multiplier]) : "$number $prefixs[$multiplier]");
}


# ------------------------------------------------------------------------------------
# nom 		: make_string_rfc822
# description 	: genere un texte valide au format des adresses email RFC822 (local part)
# parametres 	: une chaine servant de base a la generation
# retours 	: le texte genere dans tous les cas
# ------------------------------------------------------------------------------------
sub make_string_rfc822($) 
{
	my $str_base = shift;
	return 'undef' unless defined $str_base;
	conduse('FDN::Common::CharTable');

	if (ref $FDN::Common::CharTable::iso8859_15_to_rfc822 ne 'HASH') {
		return 'undef_h';
	}
	
	# On revient sur un choix de caracteres connus:
	# [a-z0-9_-]
	$str_base = cleanup_chars_as_table($str_base,
		$FDN::Common::CharTable::iso8859_15_to_rfc822);

	# nettoyage de la base
	#chomp($str_base);		# plus de \n a la fin
	$str_base =~ s/^\s*//;		# plus de blancs au debut
	$str_base =~ s/\s*$//;		# plus de blancs a la fin
	$str_base =~ s/(\W)+/$1/g;	# les caracteres non lettres et chiffres sont unifies (,,, devient ,)
	$str_base =~ s/_+/_/g;		# les underscores sont unifies (___ devient _)
	$str_base =~ s/\s+/ /g;		# les blancs consecutifs deviennent des espaces uniques

	# dcoupage de la chaine sur les mots
	my @words;
	foreach my $word (split (/ /, $str_base)) {
		# suppression des caractres non alphanumrique en dbut et fin de mot
		$word =~ s/^[\W_]*//;
		$word =~ s/[\W_]*$//;
	
		push @words, $word if (length $word);
	}

	# tous les mots trouvs dans la base sont propres, ont peut construire le login
	my $result;
	$result = join '.', @words;
	if ( length $result ) {
		return $result;
	} else {
		return 'undef';
	}
}

# ----------------------------------------------------------------------
# nom           : convert_time 
# description   : convertit une duree exprimee dans une unite en 
#                 une autre unite (minute vers seconde)
# parametres    : $time      : la dure a convertir
#                 $unite_in  : l'unite de depart
#                 $unite_out : l'untie voulue
# retour        : undef si erreur sinon la duree dans la nouvelle unite.
# ----------------------------------------------------------------------
sub convert_time($$$) {
        my ( $time, $unite_in, $unite_out) = @_;

        unless ( length($time) && $time =~ /^\d+$/ &&
                 length($unite_in) && length($unite_out)
                ){
                #error("mauvais passage de parametre");
                return undef;
        }
        # netoyage des unites
        $unite_out =~ s/^\s*(\w*)\s*$/$1/;
        $unite_in  =~ s/^\s*(\w*)\s*$/$1/;

        if ( $unite_out =~ /$unite_in/i) {
                #warning("les deux unites sont les memes");
                return $time;
        }

        my $rh_convertion = {
                'h' => { 'm' => { 'val' => 60,
                                  'op'  => 'mul'
                                },
                         's' => { 'val' => 3600,
                                  'op'  => 'mul'
                                },
                       },
                'm' => { 'h' => { 'val' => 60,
                                  'op'  => 'div'
                                },
                         's' => { 'val' => 60,
                                  'op'  => 'mul'
                                }
                       },
                's' => { 'h' => { 'val' => 3600,
                                  'op'  => 'div'
                              },
                         'm' => { 'val' => 60,
                                  'op'  => 'div'
                                }
                       }
        };
        unless( defined($rh_convertion->{lc($unite_in)}) &&
                defined($rh_convertion->{lc($unite_out)}) ) {
                error("unite inconnue");
                return undef;
        }
        my $ret_time = undef;
	$unite_in = lc $unite_in;
	$unite_out = lc $unite_out;

	my $op  = $rh_convertion->{$unite_in}->{$unite_out}->{'op'};
	my $val = $rh_convertion->{$unite_in}->{$unite_out}->{'val'};
        if( $op eq 'mul') { # on multiplie
                $ret_time = $time * $val;
        } elsif ( $op eq 'div') { # on divise
	        $ret_time = $time / $val;
		$ret_time =~ s/\..*$//;
        }
        return $ret_time;
}


# ----------------------------------------------------------------------
# nom           : make_alpha_hashing 
# description   : gnere un hash a partir d'un nom et d'une profondeur 
# parametres    : $name	 : nom
# 		  $depth : profondeur
# retour        : le hash genere.
# ----------------------------------------------------------------------
sub make_alpha_hashing($$) {
	my ( $name, $depth) = @_;

	return '' if ( $depth == 0);
	unless($name && $depth >0 && $depth =~ /^\d+$/){
		return undef;
	}

	my @base = split(//,$name);
	my ($ret_hash, $path, $char)= '';

	#for( my $i=0 ; $i < $depth ; $i++){
	
	while( $depth ) {
		$path.= $char if( $char = shift(@base)); 
		$depth--;
		$ret_hash.= '/'.$path
	} 
	return $ret_hash;
}


# ----------------------------------------------------------------------
# nom           : make_crc32_hashing 
# description   : genere un hash a partir d'un nom et d'une profondeur 
# parametres    : $name	 : nom
# 		  $depth : profondeur
# retour        : le hash genere.
# ----------------------------------------------------------------------
sub make_crc32_hashing($$) {
	my ( $name, $depth) = @_;

	return '' if ( $depth == 0 );
	unless($name && $depth >0 && $depth =~ /^\d+$/ && $depth <=4){
		return undef;
	}
	conduse('String::CRC32');
	my $crc32 = sprintf("%08lx",crc32($name)); # le 08 est il utile
	my @base;
	while($crc32) {
		$crc32 =~ s/^(..)//;
		push @base,$1; 
	}
	
	my ($ret_hash, $path, $char)= '';
	while( $depth ) {
		$path = $char if( $char = shift(@base)); 
		$depth--;
		$ret_hash.= '/'.$path
	} 
	return $ret_hash;
}

# ----------------------------------------------------------------------
# nom           : dir_hashing 
# description   : genere un hash a partir d'un nom et d'une profondeur 
# parametres    : 
#                 $type     : 
#                 $name	    : nom
# 		  $depth    : profondeur
# 		  $fallback :
# retour        : le hash genere.
# ----------------------------------------------------------------------
sub dir_hashing ($$$$) {
	my ( $type, $name, $depth, $fallback) = @_;

	if($fallback > $depth){
		$fallback = $depth;
		#FIXME peter une erreur ici
	}
	$fallback++;
	my $ra_hash_dir;
	if($type =~ /^alpha$/i){
		do {
			my $hash_dir = make_alpha_hashing($name,$depth);
			if (defined $hash_dir){
				push @{$ra_hash_dir}, $hash_dir;
			} else {
				error "impossible de generer le hash pour : ".
					"$name,$depth,$fallback";
			}
			$depth--;
			$fallback--;
		} while($fallback > 0);
	} elsif ($type =~ /^crc32$/i){
		do {
			my $hash_dir = make_crc32_hashing($name,$depth);
			if (defined $hash_dir){
				push @{$ra_hash_dir}, $hash_dir;
			} else {
				error "impossible de generer le hahs pour : ".
					"$name,$depth,$fallback";
			}
			$depth--;
			$fallback--;
		} while($fallback > 0);
	} else {
		error "type de hash inconnu";
		return undef;
	}
	return $ra_hash_dir;
}

# ----------------------------------------------------------------------
# name   : mkdir_p 
# desc   : cree les repertoire ou est stocke un fichier
# params : $newdir : path a creer
# return : 1 si OK, 0 sinon
# ----------------------------------------------------------------------
# prototype dclar avant pour ne pas pter de warning (fonction rcursive)
sub mkdir_p ($);
sub mkdir_p ($) {
	my ($newdir) = @_;
	return TRUE if -d $newdir;
	conduse('File::Basename');
	my $root = dirname($newdir);
	return FALSE if ( ! mkdir_p($root) );
	if (! mkdir $newdir) {
		error "Impossible de crer le rpertoire $newdir : $!";
		return FALSE;
	}
	return TRUE;
}



sub slurp_file($;@) {
	my ( $file, @filters ) = @_;
	local $/;
	if ( @filters ) {
		my $command = join " | ", @filters;
		open IN, "sh -c 'cat $file | $command' |";
	} else {
		open IN, "<$file" || die "Cannot read $file: $!";
	}
	my $text = <IN>;
	close IN;
	Encode::_utf8_on($text);
	if ( ! utf8::valid($text)) {
		Encode::_utf8_off($text);
		$text = Encode::decode("iso-8859-15", $text);
	}
	return $text;
}

sub slurp_binary($) {
	my ( $file ) = @_;
	local $/;
	open IN, "<$file" || die "Cannot read $file: $!";
	my $bin = <IN>;
	close IN;
	return $bin;
}

1;
